#-*- coding: utf-8 -*-
u"""

@author:		Martí Congost
@contact:		marti.congost@whads.com
@organization:	Whads/Accent SL
@since:			October 2008
"""
import cherrypy
from cocktail.pkgutils import resolve
from cocktail.events import event_handler
from cocktail.controllers import view_state
from cocktail.translations import set_language
from woost.models import get_current_user, Configuration
from woost.controllers.backoffice.basebackofficecontroller \
    import BaseBackOfficeController
from woost.controllers.backoffice.contentcontroller \
    import ContentController
from woost.controllers.backoffice.deletecontroller import DeleteController
from woost.controllers.backoffice.ordercontroller import OrderController
from woost.controllers.backoffice.movecontroller import MoveController
from woost.controllers.backoffice.renderpreviewcontroller \
    import RenderPreviewController
from woost.controllers.backoffice.editblockscontroller \
    import EditBlocksController
from woost.controllers.backoffice.dropblockcontroller \
    import DropBlockController
from woost.controllers.backoffice.changelogcontroller \
    import ChangeLogController
from woost.controllers.backoffice.uploadfilescontroller \
    import UploadFilesController
from woost.controllers.backoffice.dragandropcontroller \
    import DragAndDropController
from woost.controllers.backoffice.sitesynccontroller \
    import SiteSyncController


class BackOfficeController(BaseBackOfficeController):

    _cp_config = BaseBackOfficeController.copy_config()
    _cp_config["rendering.engine"] = "cocktail"

    _edit_stacks_manager_class = \
        "woost.controllers.backoffice.editstack.EditStacksManager"

    default_section = "content"

    content = ContentController
    delete = DeleteController
    order = OrderController
    move = MoveController
    changelog = ChangeLogController
    render_preview = RenderPreviewController
    upload_files = UploadFilesController
    drop = DragAndDropController
    blocks = EditBlocksController
    drop_block = DropBlockController

    def submit(self):
        raise cherrypy.HTTPRedirect(
            self.contextual_uri(self.default_section) + "?" + view_state())

    @event_handler
    def handle_traversed(cls, event):
        controller = event.source
        controller.context["edit_stacks_manager"] = \
            resolve(controller._edit_stacks_manager_class)()

    @event_handler
    def handle_before_request(cls, event):
        user = get_current_user()
        language = (
            user and user.prefered_language
            or Configuration.instance.get_setting("backoffice_language")
        )
        set_language(language)

    @event_handler
    def handle_after_request(cls, event):

        if event.error is None:
            controller = event.source
            edit_stacks_manager = controller.context["edit_stacks_manager"]
            edit_stack = edit_stacks_manager.current_edit_stack

            if edit_stack is not None:
                edit_stacks_manager.preserve_edit_stack(edit_stack)

    @cherrypy.expose
    def keep_alive(self, *args, **kwargs):
        pass

    synchronization = SiteSyncController




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from urllib import urlopen
import cherrypy
from woost.models import get_current_website
from woost.extensions.payments.paymentgatewaycontroller \
    import PaymentGatewayController


class DummyPaymentGatewayController(PaymentGatewayController):
    """A controller used by L{DummyPaymentGateway
    <woost.extensions.payments.dummypaymentgateway.DummyPaymentGateway>}
    instances to simulate transaction payments.
    """

    @cherrypy.expose
    def simulate_transaction(self, **parameters):

        payment_id = parameters.get("payment_id")
        payment = payment_id and self.payment_gateway.get_payment(payment_id)

        if not payment:
            raise cherrypy.NotFound()

        # Notify the payment to the application
        notification_url = self.payment_gateway.get_payment_url(
            "notification",
            payment_id = str(payment.id)
        )
        urlopen(str(notification_url))

        # Redirect the user after the transaction's over
        redirection = None

        if self.payment_gateway.payment_status == "accepted":
            redirection = self.payment_gateway.payment_successful_page
        elif self.payment_gateway.payment_status == "failed":
            redirection = self.payment_gateway.payment_failed_page

        if redirection is None:
            redirection = get_current_website().home

        raise cherrypy.HTTPRedirect(redirection.get_uri(host = "!"))




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.translations import translations
from woost.models import Extension, Role, ReadPermission

translations.define("RestrictedAccessExtension",
    ca = u"Restricció d'accés",
    es = u"Restricción de acceso",
    en = u"Access restriction"
)

translations.define("RestrictedAccessExtension-plural",
    ca = u"Restricció d'accés",
    es = u"Restricción de acceso",
    en = u"Access restriction"
)


class RestrictedAccessExtension(Extension):

    def __init__(self, **values):
        Extension.__init__(self, **values)
        self.extension_author = u"Whads/Accent SL"
        self.set("description",
            u"Facilita la definició de diferents conjunts d'elements "
            u"publicables amb accés restringit.",
            "ca"
        )
        self.set("description",
            u"Facilita la definición de diferentes conjuntos de elementos "
            u"publicables con acceso restringido.",
            "es"
        )
        self.set("description",
            u"Simplifies the definition of different sets of publishable "
            u"elements with restricted access.",
            "en"
        )

    def _load(self):

        from woost.extensions.restrictedaccess import (
            strings,
            publishable,
            accessrestriction
        )
        from woost.models import Publishable, staticpublication

        staticpublication.members_affecting_static_publication.add(
            Publishable.access_restriction
        )

    def _install(self):
        self.setup_permissions()

    def setup_permissions(self):

        everybody_role = Role.require_instance(qname = "woost.everybody")

        qname = "woost.extensions.restrictedaccess.restriction"
        restriction = ReadPermission.get_instance(qname = qname)

        if restriction is None:
            restriction = ReadPermission()
            restriction.qname = qname
            restriction.authorized = False

        restriction.content_type = Publishable
        restriction.content_expression = \
            """items.add_filter(cls.access_restriction.not_equal(""))"""
        restriction.insert()

        if restriction not in everybody_role.permissions:
            everybody_role.permissions.insert(0, restriction)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models import Item


class AudioDecoder(Item):

    instantiable = True
    visible_from_root = False

    mime_type = schema.String(
        required = True,
        unique = True,
        indexed = True,
        normalized_index = False,
        descriptive = True
    )

    command = schema.String(
        required = True
    )




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.html import Element, first_last_classes


class BlockList(Element):

    __wrap = None
    blocks = None
    container = None
    slot = None
    hide_if_empty = False
    wrap_dl_entries = False

    def __init__(self, *args, **kwargs):
        Element.__init__(self, *args, **kwargs)
        first_last_classes(self)

    def _build(self):
        self.blocks_container = self

    def _ready(self):
        Element._ready(self)

        if self.blocks is None:
            if self.container is not None and self.slot is not None:
                if isinstance(self.slot, basestring):
                    slot = self.container.__class__.get_member(self.slot)
                else:
                    slot = self.slot

                if slot:
                    self.depends_on(self.container, slot.cache_part)
                    self.blocks = getattr(self.container, slot.name, None)

        if self.tag in ("ul", "ol"):
            self.__wrap = self.wrap_with_list_item
        elif self.tag == "table":
            self.__wrap = self.wrap_with_table_row
        elif self.wrap_dl_entries and self.tag == "dl":
            self.__wrap = self.wrap_with_dl_row

        self._fill_blocks()

    def _fill_blocks(self):
        has_visible_blocks = False

        if self.blocks:
            for block in self.blocks:
                self.depends_on(block)
                if block.is_published():
                    has_visible_blocks = True
                    block_view = block.create_view()
                    self._insert_block_view(block_view)

        if self.hide_if_empty and not has_visible_blocks:
            self.visible = False

    def _insert_block_view(self, block_view):
        if self.__wrap:
            entry = self.__wrap(block_view)
            self.blocks_container.append(entry)
        else:
            self.blocks_container.append(block_view)

    def wrap_with_list_item(self, block_view):
        entry = Element("li")
        entry.append(block_view)
        return entry

    def wrap_with_table_row(self, block_view):
        row = Element("tr")
        row.cell = Element("td")
        row.cell.append(block_view)
        row.append(row.cell)
        return row

    def wrap_with_dl_row(self, block_view):
        entry = Element()
        entry.add_class("dl_entry")
        entry.append(block_view)
        return entry




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models import Website, Document


Website.add_member(
    schema.String("google_search_engine_id",
        member_group = "services.google_cse"
    )
)

Website.add_member(
    schema.Reference("google_search_results_page",
        type = Document,
        related_end = schema.Collection(),
        member_group = "services.google_cse"
    )
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
import re
from difflib import SequenceMatcher
from ZODB.broken import Broken
from cocktail.styled import styled
from cocktail import schema
from cocktail.translations import translations
from cocktail.persistence import datastore
from cocktail.persistence.utils import (
    is_broken,
    remove_broken_type as cocktail_remove_broken_type
)
from .item import Item
from .configuration import Configuration
from .permission import ContentPermission
from .changesets import Change
from .role import Role


def remove_broken_type(
    full_name,
    existing_bases = (),
    relations = (),
    excluded_relations = None,
    languages = None
):
    if languages is None:
        languages = Configuration.instance.languages

    if excluded_relations is None:
        excluded_relations = (Change.target,)

    cocktail_remove_broken_type(
        full_name,
        existing_bases = existing_bases,
        relations = relations,
        excluded_relations = excluded_relations,
        languages = languages
    )

    for role in Role.select():
        for cls in list(role.hidden_content_types):
            if (
                issubclass(cls, Broken)
                and cls.__module__ + "." + cls.__name__ == full_name
            ):
                role.hidden_content_types.remove(cls)

    for permission in ContentPermission.select():
        content_type = permission.content_type
        if content_type.__module__ + "." + content_type.__name__ == full_name:
            permission.delete()

def delete_history():

    for item in Item.select():
        if not is_broken(item):
            try:
                del item._changes
            except AttributeError:
                pass

    for key in list(datastore.root):
        if key.startswith("woost.models.changesets."):
            del datastore.root[key]

def restore_deleted_item(obj):
    if obj.insert():
        for member in obj.__class__.iter_members():
            if (
                isinstance(member, schema.Collection)
                and obj.get(member) is None
            ):
                obj.set(member, member.produce_default(instance = obj))

def iter_text(objects = None):

    if objects is None:
        objects = Item.select()

    for obj in objects:
        for member in obj.__class__.iter_members():
            if isinstance(member, schema.String):
                if member.translated:
                    languages = obj.translations.keys()
                else:
                    languages = (None,)

                for language in languages:
                    value = obj.get(member, language)
                    if value:
                        yield obj, member, language, value

def grep(expr, objects = None):

    if isinstance(expr, basestring):
        expr = re.compile(expr)

    for obj, member, language, value in iter_text(objects):
        matches = list(expr.finditer(value))
        if matches:
            yield obj, member, language, value, matches

def hl(expr, objects = None):

    for obj, member, language, value, matches in grep(expr, objects):
        print styled("-" * 100, style = "bold")
        print styled(repr(obj), style = "bold"),
        print styled(member.name, "slate_blue"),

        if language:
            print styled(language, "pink")
        else:
            print

        hl_value = value
        offset = 0

        for match in matches:
            start, end = match.span()
            original_chunk = value[start:end]
            hl_chunk = styled(original_chunk, "magenta")
            hl_value = (
                hl_value[:start + offset]
                + hl_chunk
                + hl_value[end + offset:]
            )
            offset += len(hl_chunk) - len(original_chunk)

        print hl_value

def replace(expr, replacement, objects = None, mode = "apply"):

    if isinstance(expr, basestring):
        expr = re.compile(expr)

    if mode == "apply":
        apply = True
        show = False
    elif mode == "verbose":
        apply = True
        show = True
    elif mode == "show":
        apply = False
        show = True
    elif mode == "confirm":
        apply = None
        show = True

    for obj, member, language, value in iter_text(objects):
        modified_value = expr.sub(replacement, value)

        if value == modified_value:
            continue

        if show:
            print styled("-" * 100, style = "bold")
            print styled(translations(obj), style = "bold"),
            print styled(member.name, "slate_blue"),

            if language:
                print styled(language, "pink")
            else:
                print

            diff = SequenceMatcher(a = value, b = modified_value)
            chunks = []

            for op, start_a, end_a, start_b, end_b in diff.get_opcodes():
                orig_chunk = value[start_a:end_a]
                new_chunk = modified_value[start_b:end_b]

                if op == "equal":
                    chunks.append(orig_chunk)
                else:
                    if op in ("replace", "delete"):
                        chunks.append(styled(orig_chunk, "white", "red"))

                    if op in ("replace", "insert"):
                        chunks.append(styled(new_chunk, "white", "green"))

            print "".join(chunks)

        if apply is None:
            answer = None
            while answer not in list("yn"):
                answer = raw_input("Replace (y/n)? ")
            if answer == "n":
                continue
        elif not apply:
            continue

        obj.set(member, modified_value, language)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Jordi Fernández <jordi.fernandez@whads.com>
"""
from woost.models.rendering import ImageURIRenderer
from woost.extensions.youtube.youtubevideo import YouTubeVideo

THUMBNAIL_URL_PATTERN = "http://img.youtube.com/vi/%s/0.jpg"


class YouTubeVideoRenderer(ImageURIRenderer):

    def can_render(self, item, **parameters):
        return isinstance(item, YouTubeVideo) and item.video_id

    def get_item_uri(self, item):
        return THUMBNAIL_URL_PATTERN % item.video_id




#-*- coding: utf-8 -*-
u"""Defines the `Location` class.

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail.iteration import first
from cocktail.translations import translations
from cocktail import schema
from woost.models import Item
from woost.controllers.backoffice.contentviews import global_content_views


class Location(Item):

    members_order = [
        "location_name",
        "location_type",
        "code",
        "parent",
        "locations"
    ]

    location_name = schema.String(
        translated = True,
        descriptive = True,
        required = True,
        indexed = True,
        normalized_index = True,
        full_text_indexed = True
    )

    location_type = schema.String(
        required = True,
        indexed = True,
        enumeration = [
            "continent",
            "country",
            "autonomous_community",
            "province",
            "town"
        ],
        text_search = False,
        translate_value = lambda value, language = None, **kwargs:
            "" if not value
            else translations(
                "woost.extensions.locations.location_types." + value,
                language,
                **kwargs
            ),
        edit_control = "cocktail.html.DropdownSelector"
    )

    code = schema.String(
        required = True,
        indexed = True,
        text_search = False
    )

    parent = schema.Reference(
        type = "woost.extensions.locations.location.Location",
        bidirectional = True,
        indexed = True,
        cycles_allowed = False
    )

    locations = schema.Collection(
        items = "woost.extensions.locations.location.Location",
        bidirectional = True,
        cascade_delete = True
    )

    def ascend(self, include_self = False):
        """Iterate over the location's ancestors.

        :param include_self: Indicates if the location itself should be included
            in the iteration.
        :type include_self: bool

        :return: An iterable sequence of locations.
        :rtype: `Location`
        """
        ascendant = self if include_self else self.parent
        while ascendant is not None:
            yield ascendant
            ascendant = ascendant.parent

    def descend(self, include_self = False):
        """Iterates over the location's descendants.

        :param include_self: If set to True, the location itself is yielded as
            the first element in the iteration.
        :type include_self: bool

        :return: An iterable sequence containing the location's flattened tree.
        :rtype: `Location`
        """
        if include_self:
            yield self
        for location in self.locations:
            for descendant in location.descend(True):
                yield descendant

    def descends_from(self, ascendant):
        """Indicates if the location descends from a given location.

        :param ascendant: The hypothetical container of the location.
        :type ascendant: `Location`

        :return: True if the location is, or is contained within the given
            location or one of its descendants, False otherwise.
        :rtype: bool
        """
        location = self

        while location is not None:
            if location is ascendant:
                return True
            else:
                location = location.parent

        return False

    def get_ancestor_of_type(self, location_type, include_self = False):
        """Obtains the location's first ancestor of the given type.

        :param location_type: The type of location to obtain; should match one
            of the values accepted by the `location_type` member.
        :type location_type: str

        :param include_self: If set to True, the location itself will be
            returned if it matches the indicated type. Otherwise, the location
            will be skipped and only its ancestors will be considered as
            possible matches.

        :return: The first ancestor of the location that matches the indicated
            type, or None if no such location can be found on the way to the
            tree root.
        :rtype: `Location`
        """
        location = self

        if not include_self:
            location = location.parent

        while location is not None:
            if location.location_type == location_type:
                return location
            location = location.parent

    def get_child_location(self, code, recursive = True):
        """Retrieves the contained location that matches the indicated code.

        :param code: The code of the location to retrieve.
        :type code: str

        :return: The specified location, or None if it wasn't found.
        :rtype: `Location`
        """
        for child in self.locations:
            if child.code == code:
                return child
            if recursive:
                for descendant in child.locations:
                    match = descendant.get_child_location(code)
                    if match:
                        return match

    @classmethod
    def by_code(cls, *code):

        if not code or not code[0]:
            raise ValueError(
                "Location.by_code() requires one or more location codes"
            )

        location = first(cls.select([Location.code.equal(code[0])]))

        for x in code[1:]:
            if location is None:
                return None
            location = location.get_child_location(x)

        return location

    @classmethod
    def by_path(cls, *path):

        if not path or not path[0]:
            raise ValueError(
                "Location.by_code() requires one or more location codes"
            )

        location = first(cls.select([Location.code.equal(path[0])]))

        for x in path[1:]:
            if location is None:
                return None
            location = location.get_child_location(x, recursive = False)

        return location

    def list_level(self, depth):

        if depth == 1:
            return self.locations
        else:
            descendants = []
            for location in self.locations:
                descendants.extend(location.list_level(depth - 1))
            return descendants


global_content_views.add(
    Location,
    "woost.views.TreeContentView",
    is_default = True,
    inherited = False,
    params = {
        "children_collection": Location.locations
    }
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
import os
from pkg_resources import resource_filename
from cocktail.caching import Cache


class Application(object):

    __package = None
    __root = None
    __icon_resolver = None
    __custom_icon_repository = None

    __language = None
    __authentication = None

    installation_id = None
    installation_color = None

    splash = """\
Made with                     _
__        __ ___   ___   ___ | |_
\ \  __  / // _ \ / _ \ / __||  _|
 \ \/  \/ /| (_) | (_) |___ \| |__
  \__/\__/  \___/ \___/|____/\___/

http://woost.info
"""

    def path(self, *args):
        return os.path.join(self.root, *args)

    def _get_package(self):
        return self.__package

    def _set_package(self, package):
        self.__package = package
        self.__add_custom_icon_repository()

    package = property(_get_package, _set_package)

    def _get_root(self):
        if self.__root is None and self.__package:
            return resource_filename(self.__package, "")

        return self.__root

    def _set_root(self, root):
        self.__root = root
        self.__add_custom_icon_repository()

    root = property(_get_root, _set_root)

    def __add_custom_icon_repository(self):

        root = self.root
        package = self.package

        if root and package:
            repository = (
                self.path("views", "resources", "images", "icons"),
                "/" + self.package + "_resources/images/icons"
            )

            if (
                self.__custom_icon_repository is not None
                and self.__custom_icon_repository != repository
            ):
                self.icon_resolver.icon_repositories.remove(
                    self.__custom_icon_repository
                )

            self.__custom_icon_repository = repository
            self.icon_resolver.icon_repositories.insert(0, repository)

    @property
    def icon_resolver(self):
        if self.__icon_resolver is None:
            from woost.iconresolver import IconResolver
            self.__icon_resolver = IconResolver()
        return self.__icon_resolver

    # Language scheme
    def _get_language(self):
        if self.__language is None:
            from woost.languagescheme import LanguageScheme
            self.__language = LanguageScheme()
        return self.__language

    def _set_language(self, language):
        self.__language = language

    language = property(_get_language, _set_language)

    # Authentication scheme
    def _get_authentication(self):
        if self.__authentication is None:
            from woost.authenticationscheme import AuthenticationScheme
            self.__authentication = AuthenticationScheme()
        return self.__authentication

    def _set_authentication(self, authentication):
        self.__authentication = authentication

    authentication = property(_get_authentication, _set_authentication)

    # Caching
    cache = Cache()

    # URLs
    __url_resolver = None

    def _get_url_resolver(self):
        if self.__url_resolver is None:
            from woost.urlresolver import (
                URLResolver,
                HierarchicalURLScheme,
                DescriptiveIdURLScheme
            )
            url_resolver = URLResolver()
            url_resolver.add_url_scheme(HierarchicalURLScheme())
            url_resolver.add_url_scheme(DescriptiveIdURLScheme())
            self.__url_resolver = url_resolver
        return self.__url_resolver

    def _set_url_resolver(self, url_resolver):
        self.__url_resolver = url_resolver

    url_resolver = property(_get_url_resolver, _set_url_resolver)




#-*- coding: utf-8 -*-
u"""

@author:		Martí Congost
@contact:		marti.congost@whads.com
@organization:	Whads/Accent SL
@since:			November 2013
"""
from cocktail.modeling import cached_getter
from cocktail import schema
from woost.models import User
from woost.controllers.backoffice.editstack import EditNode


class SiteInstallationEditNode(EditNode):

    @cached_getter
    def form_adapter(self):

        form_adapter = EditNode.form_adapter(self)

        if self.item.is_inserted:
            form_adapter.exclude("change_password")

            if User.encryption_method:
                kwargs = {
                    "export_condition": False,
                    "import_condition": lambda context:
                        context.get("change_password", default = None)
                }
            else:
                kwargs = {}

            form_adapter.copy("synchronization_password", **kwargs)

        return form_adapter

    @cached_getter
    def form_schema(self):

        form_schema = EditNode.form_schema(self)
        password_member = form_schema.get_member("synchronization_password")

        if password_member:

            if User.encryption_method:

                order = form_schema.members_order = list(form_schema.members_order)
                pos = order.index("synchronization_password")

                if self.item.is_inserted:

                    change_password_member = schema.Boolean(
                        name = "change_password",
                        required = True,
                        default = False,
                        visible_in_detail_view = False
                    )
                    form_schema.add_member(change_password_member)
                    order.insert(pos, "change_password")

                    password_member.exclusive = change_password_member

            # No encryption: edit passwords in plain sight
            else:
                password_member.edit_control = "cocktail.html.TextBox"

        return form_schema

    def iter_changes(self):

        # Discard differences in the password field
        for member, language in EditNode.iter_changes(self):
            if not User.encryption_method or member.name not in (
                "change_password", "synchronization_password"
            ):
                yield (member, language)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Jordi Fernández <jordi.fernandez@whads.com>
"""
from cocktail import schema
from woost.models import Configuration

Configuration.members_order.extend([
    "campaign_monitor_api_key",
    "campaign_monitor_client_id"
])

Configuration.add_member(
    schema.String("campaign_monitor_api_key",
        member_group = "services.campaign_monitor",
        listed_by_default = False
    )
)

Configuration.add_member(
    schema.String("campaign_monitor_client_id",
        member_group = "services.campaign_monitor",
        listed_by_default = False
    )
)




#-*- coding: utf-8 -*-
u"""

.. moduleauthor:: Martí Congost <marti.congost@whads.com>
"""
from cocktail import schema
from woost.models.rendering.renderer import Renderer


class ChainRenderer(Renderer):

    instantiable = True

    renderers = schema.Collection(
        items = schema.Reference(type = Renderer),
        bidirectional = True,
        integral = True,
        related_end = schema.Reference()
    )

    def can_render(self, item):
        return any(renderer.can_render(item) for renderer in self.renderers)

    def render(self, item, **parameters):
        for renderer in self.renderers:
            if renderer.can_render(item, **parameters):
                return renderer.render(item, **parameters)


